"""Models specific to the resource strategy."""
from typing import Any, Dict, List, Literal, Optional

from oteapi.models import ResourceConfig, SessionUpdate
from pydantic import Field

from oteapi_optimade.models.config import OPTIMADEConfig
from oteapi_optimade.models.custom_types import OPTIMADEUrl


class OPTIMADEResourceConfig(ResourceConfig):
    """OPTIMADE-specific resource strategy config."""

    accessUrl: OPTIMADEUrl = Field(
        ...,
        description="Either a base OPTIMADE URL or a full OPTIMADE URL.",
    )
    accessService: Literal["optimade", "OPTIMADE", "OPTiMaDe"] = Field(
        ...,
        description="The registered strategy name for OPTIMADEResourceStrategy.",
    )
    configuration: OPTIMADEConfig = Field(
        OPTIMADEConfig(),
        description=(
            "OPTIMADE configuration. Contains relevant information necessary to "
            "perform OPTIMADE queries."
        ),
    )


class OPTIMADEResourceSession(SessionUpdate):
    """OPTIMADE session for the resource strategy."""

    optimade_config: Optional[OPTIMADEConfig] = Field(
        None,
        description=(
            "OPTIMADE configuration. Contains relevant information necessary to "
            "perform OPTIMADE queries."
        ),
    )
    optimade_resources: List[Dict[str, Any]] = Field(
        [],
        description=(
            "List of OPTIMADE resources (structures, references, errors, ...) returned"
            " from the OPTIMADE request."
        ),
    )
    optimade_resource_model: str = Field(
        "",
        description=(
            "Importable path to the resource model to be used to parse the OPTIMADE "
            "resources in `optimade_resource`. The importable path should be a fully "
            "importable path to a module separated by a colon (`:`) to then define the "
            "resource model class name. This means one can then do:\n\n```python\n"
            "from PACKAGE.MODULE import RESOURCE_CLS\n```\nFrom the value "
            "`PACKAGE.MODULE:RESOURCE_CLS`"
        ),
        regex=(
            r"^([a-zA-Z][a-zA-Z0-9_]*(\.[a-zA-Z][a-zA-Z0-9_]*)*"  # package.module
            r":[a-zA-Z][a-zA-Z0-9_]*)?$"  # class
        ),
    )

    class Config:
        """Pydantic configuration for `OPTIMADEResourceSession`."""

        validate_assignment = True
        arbitrary_types_allowed = True



"""Test `oteapi_optimade.strategies.resource` module.
Specifically the `OPTIMADEResourceStrategy` class.
"""
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Optional

    from requests_mock import Mocker


@pytest.fixture
def resource_config() -> dict[str, str]:
    """A resource config dictionary for test purposes."""
    return {
        "accessService": "optimade",
        "accessUrl": (
            "https://example.org/some/base/v0.1/optimade/v1/structures"
            '?filter=elements HAS ALL "Si","O"&sort=nelements&page_limit=2'
        ),
    }


@pytest.mark.parametrize(
    "session", [None, {"optimade_config": "content"}], ids=["None", "dict"]
)
def test_initialize(session: "Optional[dict]", resource_config: dict[str, str]) -> None:
    """Test the `initialize()` method."""
    from oteapi.models import SessionUpdate

    from oteapi_optimade.strategies.resource import OPTIMADEResourceStrategy

    session_update = OPTIMADEResourceStrategy(resource_config).initialize(session)

    assert isinstance(session_update, SessionUpdate)
    assert {**session_update} == {}


def test_get_no_session(
    resource_config: dict[str, str], static_files: "Path", requests_mock: "Mocker"
) -> None:
    """Test the `get()` method - session is `None`."""
    from optimade.adapters import Structure

    from oteapi_optimade.models.strategies.resource import OPTIMADEResourceSession
    from oteapi_optimade.strategies.resource import OPTIMADEResourceStrategy

    sample_file = static_files / "optimade_response.json"
    requests_mock.get(resource_config["accessUrl"], content=sample_file.read_bytes())

    session = OPTIMADEResourceStrategy(resource_config).get()

    assert isinstance(session, OPTIMADEResourceSession)
    assert session.optimade_config is None
    assert session.optimade_resource_model == f"{Structure.__module__}:Structure"
    assert session.optimade_resources
    for resource in session.optimade_resources:
        assert Structure(resource)



"""Pytest fixtures and configuration."""
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from pathlib import Path


def pytest_configure(config) -> None:  # pylint: disable=unused-argument
    """Method that runs before pytest collects tests, so no modules are imported."""
    import os
    from pathlib import Path

    cwd = Path(__file__).resolve().parent.resolve()
    os.environ["OPTIMADE_CONFIG_FILE"] = str(
        cwd / "static" / "test_optimade_config.yml"
    )


@pytest.fixture(scope="session")
def top_dir() -> "Path":
    """Return resolved Path object to the repository's top directory."""
    from pathlib import Path

    return Path(__file__).resolve().parent.parent.resolve()


@pytest.fixture(scope="session")
def static_files(top_dir: "Path") -> "Path":
    """Return Path object to the `static` folder."""
    return (top_dir / "tests" / "static").resolve()



"""Pytest fixtures for `strategies/`."""
import pytest


@pytest.fixture(scope="session", autouse=True)
def load_strategies() -> None:
    """Load entry points strategies."""
    from oteapi.plugins import load_strategies

    load_strategies(test_for_uniqueness=False)



"""Data models related to OPTIMADE queries."""
from typing import Optional
from urllib.parse import quote, unquote, urlencode

from optimade.server.query_params import EntryListingQueryParams
from pydantic import BaseModel, EmailStr, Field

QUERY_PARAMETERS = EntryListingQueryParams()
"""Entry listing URL query parameters from the `optimade` package
([`EntryListingQueryParams`](https://www.optimade.org/optimade-python-tools/api_reference/server/query_params/#optimade.server.query_params.EntryListingQueryParams))."""  # pylint: disable=line-too-long


class OPTIMADEQueryParameters(BaseModel, validate_assignment=True):
    """Common OPTIMADE entry listing endpoint query parameters."""

    filter: Optional[str] = Field(
        QUERY_PARAMETERS.filter.default,
        description=QUERY_PARAMETERS.filter.description,
    )
    response_format: Optional[str] = Field(
        QUERY_PARAMETERS.response_format.default,
        description=QUERY_PARAMETERS.response_format.description,
    )
    email_address: Optional[EmailStr] = Field(
        QUERY_PARAMETERS.email_address.default,
        description=QUERY_PARAMETERS.email_address.description,
    )
    response_fields: Optional[str] = Field(
        QUERY_PARAMETERS.response_fields.default,
        description=QUERY_PARAMETERS.response_fields.description,
        regex=QUERY_PARAMETERS.response_fields.regex,
    )
    sort: Optional[str] = Field(
        QUERY_PARAMETERS.sort.default,
        description=QUERY_PARAMETERS.sort.description,
        regex=QUERY_PARAMETERS.sort.regex,
    )
    page_limit: Optional[int] = Field(
        QUERY_PARAMETERS.page_limit.default,
        description=QUERY_PARAMETERS.page_limit.description,
        ge=QUERY_PARAMETERS.page_limit.ge,
    )
    page_offset: Optional[int] = Field(
        QUERY_PARAMETERS.page_offset.default,
        description=QUERY_PARAMETERS.page_offset.description,
        ge=QUERY_PARAMETERS.page_offset.ge,
    )
    page_number: Optional[int] = Field(
        QUERY_PARAMETERS.page_number.default,
        description=QUERY_PARAMETERS.page_number.description,
        ge=QUERY_PARAMETERS.page_number.ge,
    )
    page_cursor: Optional[int] = Field(
        QUERY_PARAMETERS.page_cursor.default,
        description=QUERY_PARAMETERS.page_cursor.description,
        ge=QUERY_PARAMETERS.page_cursor.ge,
    )
    page_above: Optional[int] = Field(
        QUERY_PARAMETERS.page_above.default,
        description=QUERY_PARAMETERS.page_above.description,
        ge=QUERY_PARAMETERS.page_above.ge,
    )
    page_below: Optional[int] = Field(
        QUERY_PARAMETERS.page_below.default,
        description=QUERY_PARAMETERS.page_below.description,
        ge=QUERY_PARAMETERS.page_below.ge,
    )
    include: Optional[str] = Field(
        QUERY_PARAMETERS.include.default,
        description=QUERY_PARAMETERS.include.description,
    )
    # api_hint is not yet initialized in `EntryListingQueryParams`.
    # These values are copied verbatim from `optimade==0.16.10`.
    api_hint: Optional[str] = Field(
        "",
        description=(
            "If the client provides the parameter, the value SHOULD have the format "
            "`vMAJOR` or `vMAJOR.MINOR`, where MAJOR is a major version and MINOR is a"
            " minor version of the API. For example, if a client appends "
            "`api_hint=v1.0` to the query string, the hint provided is for major "
            "version 1 and minor version 0."
        ),
        regex=r"(v[0-9]+(\.[0-9]+)?)?",
    )

    def generate_query_string(self) -> str:
        """Generate a valid URL query string based on the set fields."""
        res = {}
        for field, value in self.dict().items():
            if (
                value
                or field
                in self.__fields_set__  # pylint: disable=unsupported-membership-test
            ):
                res[field] = unquote(value) if isinstance(value, str) else value
        return urlencode(res, quote_via=quote)



"""OPTIMADE resource strategy."""
import logging
import sys
from typing import TYPE_CHECKING
from urllib.parse import parse_qs

import requests
from optimade.adapters import Reference, Structure
from optimade.models import (
    ErrorResponse,
    OptimadeError,
    ReferenceResponseMany,
    ReferenceResponseOne,
    StructureResponseMany,
    StructureResponseOne,
)
from oteapi.datacache import DataCache
from oteapi.models import SessionUpdate
from oteapi.plugins import create_strategy
from oteapi.plugins.entry_points import StrategyType
from pydantic.dataclasses import dataclass

from oteapi_optimade.exceptions import OPTIMADEParseError
from oteapi_optimade.models import OPTIMADEResourceConfig, OPTIMADEResourceSession
from oteapi_optimade.models.custom_types import OPTIMADEUrl
from oteapi_optimade.models.query import OPTIMADEQueryParameters
from oteapi_optimade.utils import model2dict

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, Union

    from optimade.models import Response as OPTIMADEResponse


LOGGER = logging.getLogger("oteapi_optimade.strategies")
LOGGER.setLevel(logging.DEBUG)
LOGGER.addHandler(logging.StreamHandler(sys.stdout))


@dataclass
class OPTIMADEResourceStrategy:
    """OPTIMADE Resource Strategy.

    **Implements strategies**:

    - `("accessService", "optimade")`
    - `("accessService", "OPTIMADE")`
    - `("accessService", "OPTiMaDe")`

    """

    resource_config: OPTIMADEResourceConfig

    def initialize(  # pylint: disable=unused-argument
        self, session: "Optional[Dict[str, Any]]" = None
    ) -> SessionUpdate:
        """Initialize strategy.

        This method will be called through the `/initialize` endpoint of the OTE-API
        Services.

        Parameters:
            session: A session-specific dictionary context.

        Returns:
            An update model of key/value-pairs to be stored in the session-specific
            context from services.

        """
        return SessionUpdate()

    def get(  # pylint: disable=too-many-branches,too-many-statements
        self, session: "Optional[Union[SessionUpdate, Dict[str, Any]]]" = None
    ) -> OPTIMADEResourceSession:
        """Execute an OPTIMADE query to `accessUrl`.

        This method will be called through the strategy-specific endpoint of the
        OTE-API Services.

        Configuration values provided in `resource_config.configuration` take
        precedence over the derived values from `accessUrl`.

        Workflow:
        1. Update configuration according to session.
        2. Deconstruct `accessUrl` (done partly by
           `oteapi_optimade.models.custom_types.OPTIMADEUrl`).
        3. Reconstruct the complete query URL.
        4. Send query.
        5. Store result in data cache.

        Parameters:
            session: A session-specific dictionary-like context.

        Returns:
            An update model of key/value-pairs to be stored in the session-specific
            context from services.

        """
        if session and isinstance(session, dict):
            session = OPTIMADEResourceSession(**session)
        elif session and isinstance(session, SessionUpdate):
            session = OPTIMADEResourceSession(
                **model2dict(session, exclude_defaults=True, exclude_unset=True)
            )
        else:
            session = OPTIMADEResourceSession()

        if session.optimade_config:
            self.resource_config.configuration.update(
                model2dict(
                    session.optimade_config, exclude_defaults=True, exclude_unset=True
                )
            )

        optimade_endpoint = self.resource_config.accessUrl.endpoint or "structures"
        optimade_query = (
            self.resource_config.configuration.query_parameters
            or OPTIMADEQueryParameters()
        )
        LOGGER.debug("resource_config: %r", self.resource_config)

        if self.resource_config.accessUrl.query:
            parsed_query = parse_qs(self.resource_config.accessUrl.query)
            for field, value in parsed_query.items():
                # Only use the latest defined value for any parameter
                if field not in optimade_query.__fields_set__:
                    LOGGER.debug(
                        "Setting %r from accessUrl (value=%r)", field, value[-1]
                    )
                    setattr(optimade_query, field, value[-1])

        LOGGER.debug("optimade_query after update: %r", optimade_query)

        optimade_url = OPTIMADEUrl(
            f"{self.resource_config.accessUrl.base_url}"
            f"{'/' + self.resource_config.accessUrl.version if self.resource_config.accessUrl.version else '/v1'}"  # pylint: disable=line-too-long
            f"/{optimade_endpoint}?{optimade_query.generate_query_string()}"
        )
        LOGGER.debug("OPTIMADE URL to be requested: %s", optimade_url)

        # Set cache access key to the full OPTIMADE URL.
        self.resource_config.configuration.datacache_config.accessKey = optimade_url

        # Perform query
        response = requests.get(optimade_url, allow_redirects=True)

        if optimade_query.response_format and optimade_query.response_format != "json":
            raise NotImplementedError(
                "Can only handle JSON responses for now. Requested response format: "
                f"{optimade_query.response_format!r}"
            )

        cache = DataCache(config=self.resource_config.configuration.datacache_config)
        cache.add(
            {
                "status_code": response.status_code,
                "ok": response.ok,
                "json": response.json(),
            }
        )

        parse_config = {
            "downloadUrl": optimade_url,
            "mediaType": (
                f"application/vnd.{self.resource_config.accessService}"
                f"{'+' + optimade_query.response_format if optimade_query.response_format else ''}"  # pylint: disable=line-too-long
            ),
            "configuration": {
                "datacache_config": self.resource_config.configuration.datacache_config,
                "return_object": True,
            },
        }

        session.update(
            create_strategy(StrategyType.PARSE, parse_config).initialize(
                model2dict(session, exclude_defaults=True, exclude_unset=True)
            )
        )
        session.update(
            create_strategy(StrategyType.PARSE, parse_config).get(
                model2dict(session, exclude_defaults=True, exclude_unset=True)
            )
        )

        if "optimade_response_object" not in session:
            raise ValueError(
                "'optimade_response_object' was expected to be present in the session."
            )
        optimade_response: "OPTIMADEResponse" = session.pop("optimade_response_object")
        if "optimade_response" in session and not session.get("optimade_response"):
            del session["optimade_response"]

        if isinstance(optimade_response, ErrorResponse):
            optimade_resources = optimade_response.errors
            session.optimade_resource_model = (
                f"{OptimadeError.__module__}:OptimadeError"
            )
        elif isinstance(optimade_response, ReferenceResponseMany):
            optimade_resources = [
                Reference(entry).as_dict
                if isinstance(entry, dict)
                else Reference(entry.dict()).as_dict
                for entry in optimade_response.data
            ]
            session.optimade_resource_model = f"{Reference.__module__}:Reference"
        elif isinstance(optimade_response, ReferenceResponseOne):
            optimade_resources = [
                Reference(optimade_response.data).as_dict
                if isinstance(optimade_response.data, dict)
                else Reference(optimade_response.data.dict()).as_dict
            ]
            session.optimade_resource_model = f"{Reference.__module__}:Reference"
        elif isinstance(optimade_response, StructureResponseMany):
            optimade_resources = [
                Structure(entry).as_dict
                if isinstance(entry, dict)
                else Structure(entry.dict()).as_dict
                for entry in optimade_response.data
            ]
            session.optimade_resource_model = f"{Structure.__module__}:Structure"
        elif isinstance(optimade_response, StructureResponseOne):
            optimade_resources = [
                Structure(optimade_response.data).as_dict
                if isinstance(optimade_response.data, dict)
                else Structure(optimade_response.data.dict()).as_dict
            ]
            session.optimade_resource_model = f"{Structure.__module__}:Structure"
        else:
            LOGGER.debug(
                "Could not parse response as errors, references or structures. "
                "Response:\n%r",
                optimade_response,
            )
            raise OPTIMADEParseError(
                "Could not retrieve errors, references or structures from response "
                f"from {optimade_url}. It could be a valid OPTIMADE API response, "
                "however it may not be supported by OTEAPI-OPTIMADE. It may also be an "
                "invalid response completely."
            )

        session.optimade_resources = [
            model2dict(resource) for resource in optimade_resources
        ]

        if session.optimade_config and session.optimade_config.query_parameters:
            session = session.copy(
                update={
                    "optimade_config": session.optimade_config.copy(
                        update={
                            "query_parameters": model2dict(
                                session.optimade_config.query_parameters,
                                exclude_defaults=True,
                                exclude_unset=True,
                            )
                        }
                    )
                }
            )

        return session



"""OTE-API OPTIMADE-specific Python exceptions."""


class BaseOteapiOptimadeException(Exception):
    """Base OTE-API OPTIMADE exception."""


class ConfigurationError(BaseOteapiOptimadeException):
    """An error occurred when dealing with strategy configurations."""


class RequestError(BaseOteapiOptimadeException):
    """A general error occured when performing a URL request."""


class OPTIMADEResponseError(RequestError):
    """An OPTIMADE error was returned from a URL request."""


class OPTIMADEParseError(BaseOteapiOptimadeException):
    """Could not use OPTIMADE Python tools to parse an OPTIMADE API response."""



"""`oteapi_optimade.models` module - pydantic data models."""
from .strategies import (
    OPTIMADEFilterConfig,
    OPTIMADEFilterSession,
    OPTIMADEParseConfig,
    OPTIMADEParseSession,
    OPTIMADEResourceConfig,
    OPTIMADEResourceSession,
)

__all__ = (
    "OPTIMADEFilterConfig",
    "OPTIMADEFilterSession",
    "OPTIMADEParseConfig",
    "OPTIMADEParseSession",
    "OPTIMADEResourceConfig",
    "OPTIMADEResourceSession",
)



"""Demo strategy class for text/json."""
import json
import logging
import sys
from typing import TYPE_CHECKING

from optimade.models import ErrorResponse, Success
from oteapi.datacache import DataCache
from oteapi.models import SessionUpdate
from oteapi.plugins import create_strategy
from oteapi.plugins.entry_points import StrategyType
from pydantic import ValidationError
from pydantic.dataclasses import dataclass

from oteapi_optimade.exceptions import OPTIMADEParseError
from oteapi_optimade.models import OPTIMADEParseConfig, OPTIMADEParseSession
from oteapi_optimade.utils import model2dict

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, Union


LOGGER = logging.getLogger("oteapi_optimade.strategies")
LOGGER.setLevel(logging.DEBUG)
LOGGER.addHandler(logging.StreamHandler(sys.stdout))


@dataclass
class OPTIMADEParseStrategy:
    """Parse strategy for JSON.

    **Implements strategies**:

    - `("mediaType", "application/vnd.optimade+json")`
    - `("mediaType", "application/vnd.OPTIMADE+json")`
    - `("mediaType", "application/vnd.OPTiMaDe+json")`
    - `("mediaType", "application/vnd.optimade+JSON")`
    - `("mediaType", "application/vnd.OPTIMADE+JSON")`
    - `("mediaType", "application/vnd.OPTiMaDe+JSON")`
    - `("mediaType", "application/vnd.optimade")`
    - `("mediaType", "application/vnd.OPTIMADE")`
    - `("mediaType", "application/vnd.OPTiMaDe")`

    """

    parse_config: OPTIMADEParseConfig

    def initialize(  # pylint: disable=unused-argument
        self, session: "Optional[Dict[str, Any]]" = None
    ) -> SessionUpdate:
        """Initialize strategy.

        This method will be called through the `/initialize` endpoint of the OTE-API
        Services.

        Parameters:
            session: A session-specific dictionary context.

        Returns:
            An update model of key/value-pairs to be stored in the session-specific
            context from services.

        """
        return SessionUpdate()

    def get(  # pylint: disable=too-many-branches
        self, session: "Optional[Union[SessionUpdate, Dict[str, Any]]]" = None
    ) -> OPTIMADEParseSession:
        """Request and parse an OPTIMADE response using OPT.

        This method will be called through the strategy-specific endpoint of the
        OTE-API Services.

        Configuration values provided in `resource_config.configuration` take
        precedence over the derived values from `downloadUrl`.

        Workflow:
        1. Request OPTIMADE response.
        2. Parse as an OPTIMADE Python tools (OPT) pydantic response model.

        Parameters:
            session: A session-specific dictionary-like context.

        Returns:
            An update model of key/value-pairs to be stored in the session-specific
            context from services.

        """
        if session and isinstance(session, dict):
            session = OPTIMADEParseSession(**session)
        elif session and isinstance(session, SessionUpdate):
            session = OPTIMADEParseSession(
                **model2dict(session, exclude_defaults=True, exclude_unset=True)
            )
        else:
            session = OPTIMADEParseSession()

        if session.optimade_config:
            self.parse_config.configuration.update(
                model2dict(
                    session.optimade_config, exclude_defaults=True, exclude_unset=True
                )
            )

        cache = DataCache(self.parse_config.configuration.datacache_config)
        if self.parse_config.downloadUrl in cache:
            response: "Dict[str, Any]" = cache.get(self.parse_config.downloadUrl)
        elif (
            self.parse_config.configuration.datacache_config.accessKey
            and self.parse_config.configuration.datacache_config.accessKey in cache
        ):
            response = cache.get(
                self.parse_config.configuration.datacache_config.accessKey
            )
        else:
            download_config = self.parse_config.copy()
            session.update(
                create_strategy(StrategyType.DOWNLOAD, download_config).initialize(
                    model2dict(session, exclude_defaults=True, exclude_unset=True)
                )
            )
            session.update(
                create_strategy(StrategyType.DOWNLOAD, download_config).get(
                    model2dict(session, exclude_defaults=True, exclude_unset=True)
                )
            )

            response = {"json": json.loads(cache.get(session.pop("key")))}

        if (
            not response.get("ok", True)
            or (
                200 > response.get("status_code", 200)
                or response.get("status_code", 200) >= 300
            )
            or "errors" in response.get("json", {})
        ):
            # Error response
            try:
                response_object = ErrorResponse(**response.get("json", {}))
            except ValidationError as exc:
                LOGGER.error(
                    "Could not validate an error response.\nValidationError: "
                    "%s\nresponse=%r",
                    exc,
                    response,
                )
                raise OPTIMADEParseError(
                    "Could not validate an error response."
                ) from exc
        else:
            # Successful response
            response_model = self.parse_config.downloadUrl.response_model()
            if response_model:
                if not isinstance(response_model, tuple):
                    response_model = (response_model,)
                for model_cls in response_model:
                    try:
                        response_object = model_cls(**response.get("json", {}))
                    except ValidationError:
                        pass
                    else:
                        break
                else:
                    LOGGER.error(
                        "Could not validate for an expected response model.\nURL=%r\n"
                        "response_models=%r\nresponse=%s",
                        self.parse_config.downloadUrl,
                        response_model,
                        response,
                    )
                    raise OPTIMADEParseError(
                        "Could not validate for an expected response model."
                    )
            else:
                # No "endpoint" or unknown
                try:
                    response_object = Success(**response.get("json", {}))
                except ValidationError as exc:
                    LOGGER.error(
                        "Unknown or unparseable endpoint.\nValidatonError: %s\n"
                        "URL=%r\nendpoint=%r\nresponse_model=%r\nresponse=%s",
                        exc,
                        self.parse_config.downloadUrl,
                        self.parse_config.downloadUrl.endpoint,
                        response_model,
                        response,
                    )
                    raise OPTIMADEParseError(
                        "Unknown or unparseable endpoint."
                    ) from exc

        if self.parse_config.configuration.return_object:
            session.optimade_response_object = response_object
        else:
            session.optimade_response = model2dict(response_object)

        if session.optimade_config and session.optimade_config.query_parameters:
            session = session.copy(
                update={
                    "optimade_config": session.optimade_config.copy(
                        update={
                            "query_parameters": model2dict(
                                session.optimade_config.query_parameters,
                                exclude_defaults=True,
                                exclude_unset=True,
                            )
                        }
                    )
                }
            )

        return session



"""Demo filter strategy."""
import logging
import sys
from typing import TYPE_CHECKING

from oteapi.models import SessionUpdate
from pydantic.dataclasses import dataclass

from oteapi_optimade.models import OPTIMADEFilterConfig, OPTIMADEFilterSession
from oteapi_optimade.models.query import OPTIMADEQueryParameters
from oteapi_optimade.utils import model2dict

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, Union


LOGGER = logging.getLogger("oteapi_optimade.strategies")
LOGGER.setLevel(logging.DEBUG)
LOGGER.addHandler(logging.StreamHandler(sys.stdout))


@dataclass
class OPTIMADEFilterStrategy:
    """Filter Strategy.

    **Implements strategies**:

    - `("filterType", "OPTIMADE")`
    - `("filterType", "optimade")`
    - `("filterType", "OPTiMaDe")`

    """

    filter_config: OPTIMADEFilterConfig

    def initialize(
        self, session: "Optional[Union[SessionUpdate, Dict[str, Any]]]" = None
    ) -> OPTIMADEFilterSession:
        """Initialize strategy.

        This method will be called through the `/initialize` endpoint of the OTE-API
        Services.

        Configuration values, specifically URL query parameters, can be provided to the
        OPTIMADE resource strategy through this filter strategy.

        Workflow:
        1. Compile received information.
        2. Update session with compiled information.

        Parameters:
            session: A session-specific dictionary context.

        Returns:
            An update model of key/value-pairs to be stored in the
            session-specific context from services.

        """
        if session and isinstance(session, dict):
            session = OPTIMADEFilterSession(**session)
        elif session and isinstance(session, SessionUpdate):
            session = OPTIMADEFilterSession(
                **model2dict(session, exclude_defaults=True, exclude_unset=True)
            )
        else:
            session = OPTIMADEFilterSession()

        if session.optimade_config:
            self.filter_config.configuration.update(
                model2dict(
                    session.optimade_config, exclude_defaults=True, exclude_unset=True
                )
            )

        optimade_config = self.filter_config.configuration.copy()

        if not optimade_config.query_parameters:
            optimade_config.query_parameters = OPTIMADEQueryParameters()

        if self.filter_config.query:
            LOGGER.debug("Setting filter from query.")
            optimade_config.query_parameters.filter = self.filter_config.query

        if self.filter_config.limit:
            LOGGER.debug("Setting page_limit from limit.")
            optimade_config.query_parameters.page_limit = self.filter_config.limit

        return session.copy(
            update={
                "optimade_config": optimade_config.copy(
                    update={
                        "query_parameters": model2dict(
                            optimade_config.query_parameters,
                            exclude_defaults=True,
                            exclude_unset=True,
                        )
                    }
                )
            },
        )

    def get(  # pylint: disable=unused-argument
        self,
        session: "Optional[Dict[str, Any]]" = None,
    ) -> SessionUpdate:
        """Execute the strategy.

        This method will be called through the strategy-specific endpoint of the
        OTE-API Services.

        Parameters:
            session: A session-specific dictionary context.

        Returns:
            An update model of key/value-pairs to be stored in the
            session-specific context from services.

        """
        return SessionUpdate()



"""OTE-API OPTIMADE

A plugin for OTE-API.

Authored by Casper Welzel Andersen, SINTEF, 2022
Created from cookiecutter-oteapi-plugin, SINTEF, 2022
"""

__version__ = "0.2.2"
__author__ = "Casper Welzel Andersen"
__author_email__ = "casper.w.andersen@sintef.no"



"""General OPTIMADE configuration models."""
from typing import Literal, Optional

from oteapi.models import AttrDict, DataCacheConfig
from pydantic import Field, validator

from oteapi_optimade.models.query import OPTIMADEQueryParameters

DEFAULT_CACHE_CONFIG_VALUES = {
    "expireTime": 60 * 60 * 24,  # 1 day (in seconds) = 60 s/min * 60 min/h * 24 h/day
    "tag": "optimade",
}
"""Set the `expireTime` and `tag` to default values for the data cache."""


class OPTIMADEConfig(AttrDict):
    """OPTIMADE configuration."""

    version: str = Field(
        "v1",
        description="The version part of the OPTIMADE versioned base URL.",
        regex=r"^v[0-9]+(\.[0-9]+){,2}$",
    )
    endpoint: Literal["references", "structures"] = Field(
        "structures",
        description="Supported OPTIMADE entry resource endpoint.",
    )
    query_parameters: Optional[OPTIMADEQueryParameters] = Field(
        None,
        description="URL query parameters to be used in the OPTIMADE query.",
    )
    datacache_config: DataCacheConfig = Field(
        DataCacheConfig(**DEFAULT_CACHE_CONFIG_VALUES),
        description="Configuration options for the local data cache.",
    )
    return_object: bool = Field(
        False,
        description=(
            "Whether or not to return a response object (using the pydantic model).\n"
            "\nImportant:\n    This should _only_ be used if the strategy is called "
            "directly and not via an OTEAPI REST API service."
        ),
    )

    @validator("datacache_config")
    def default_datacache_config(cls, value: DataCacheConfig) -> DataCacheConfig:
        """Use default values for `DataCacheConfig` if not supplied."""
        original_set_values = len(value.__fields_set__)

        for field, default_value in DEFAULT_CACHE_CONFIG_VALUES.items():
            if field in value.__fields_set__:
                # Use the set value instead of the default
                continue
            setattr(value, field, default_value)

        if len(value.__fields_set__) > original_set_values:
            # Re-validate model and return it
            return value.validate(
                {
                    field: field_value
                    for field, field_value in value.dict().items()
                    if field in value.__fields_set__
                }
            )
        return value



"""Models specific to the filter strategy."""
from typing import Any, Dict, Literal, Optional

from optimade.models import Response
from oteapi.models import FilterConfig, SessionUpdate
from pydantic import Field

from oteapi_optimade.models.config import OPTIMADEConfig


class OPTIMADEFilterConfig(FilterConfig):
    """OPTIMADE-specific filter strategy config.

    Note:
        The `condition` parameter is not taken into account.

    """

    filterType: Literal["optimade", "OPTIMADE", "OPTiMaDe"] = Field(
        ...,
        description="The registered strategy name for OPTIMADEFilterStrategy.",
    )
    query: Optional[str] = Field(
        None,
        description=(
            "The `filter` OPTIMADE query parameter value. This parameter value can "
            "also be provided through the [`configuration.query_parameters.filter`]"
            "[oteapi_optimade.models.query.OPTIMADEQueryParameters.filter] parameter. "
            "Note, this value takes precedence over [`configuration`][oteapi_optimade."
            "models.strategies.filter.OPTIMADEFilterConfig.configuration] values."
        ),
    )
    limit: Optional[int] = Field(
        None,
        description=(
            "The `page_limit` OPTIMADE query parameter value. This parameter value can"
            " also be provided through the [`configuration.query_parameters."
            "page_limit`][oteapi_optimade.models.query.OPTIMADEQueryParameters."
            "page_limit] parameter. Note, this value takes precedence over "
            "[`configuration`][oteapi_optimade.models.strategies.filter."
            "OPTIMADEFilterConfig.configuration] values."
        ),
    )
    configuration: OPTIMADEConfig = Field(
        OPTIMADEConfig(),
        description=(
            "OPTIMADE configuration. Contains relevant information necessary to "
            "perform OPTIMADE queries."
        ),
    )


class OPTIMADEFilterSession(SessionUpdate):
    """OPTIMADE session for the filter strategy."""

    optimade_config: Optional[OPTIMADEConfig] = Field(
        None,
        description=(
            "OPTIMADE configuration. Contains relevant information necessary to "
            "perform OPTIMADE queries."
        ),
    )
    optimade_response_object: Optional[Response] = Field(
        None,
        description="An OPTIMADE Python tools (OPT) pydantic response object.",
    )
    optimade_response: Optional[Dict[str, Any]] = Field(
        None,
        description="An OPTIMADE response as a Python dictionary.",
    )

    class Config:
        """Pydantic configuration for `OPTIMADEFilterSession`."""

        validate_assignment = True
        arbitrary_types_allowed = True



"""Models specific to the parse strategy."""
from typing import Any, Dict, Literal, Optional

from optimade.models import Response
from oteapi.models import ResourceConfig, SessionUpdate
from pydantic import Field

from oteapi_optimade.models.config import OPTIMADEConfig
from oteapi_optimade.models.custom_types import OPTIMADEUrl


class OPTIMADEParseConfig(ResourceConfig):
    """OPTIMADE-specific parse strategy config."""

    downloadUrl: OPTIMADEUrl = Field(
        ...,
        description="Either a base OPTIMADE URL or a full OPTIMADE URL.",
    )
    mediaType: Literal[
        "application/vnd.optimade+json",
        "application/vnd.OPTIMADE+json",
        "application/vnd.OPTiMaDe+json",
        "application/vnd.optimade+JSON",
        "application/vnd.OPTIMADE+JSON",
        "application/vnd.OPTiMaDe+JSON",
        "application/vnd.optimade",
        "application/vnd.OPTIMADE",
        "application/vnd.OPTiMaDe",
    ] = Field(
        ...,
        description="The registered strategy name for OPTIMADEParseStrategy.",
    )
    configuration: OPTIMADEConfig = Field(
        OPTIMADEConfig(),
        description=(
            "OPTIMADE configuration. Contains relevant information necessary to "
            "perform OPTIMADE queries."
        ),
    )


class OPTIMADEParseSession(SessionUpdate):
    """OPTIMADE session for the parse strategy."""

    optimade_config: Optional[OPTIMADEConfig] = Field(
        None,
        description=(
            "OPTIMADE configuration. Contains relevant information necessary to "
            "perform OPTIMADE queries."
        ),
    )
    optimade_response_object: Optional[Response] = Field(
        None,
        description="An OPTIMADE Python tools (OPT) pydantic response object.",
    )
    optimade_response: Optional[Dict[str, Any]] = Field(
        None,
        description="An OPTIMADE response as a Python dictionary.",
    )

    class Config:
        """Pydantic configuration for `OPTIMADEParseSession`."""

        validate_assignment = True
        arbitrary_types_allowed = True



"""Utility functions to be used in OTEAPI OPTIMADE."""
from copy import deepcopy
from typing import TYPE_CHECKING, Iterable

from pydantic import BaseModel

if TYPE_CHECKING:
    from typing import Any, Union


def model2dict(
    model: "Union[BaseModel, dict[str, Any]]", **dict_kwargs: "Any"
) -> "dict[str, Any]":
    """Convert a pydantic model to a Python dictionary.

    This works similarly to the `dict()` method for pydantic models, but ensures any
    and all nested pydantic models are also converted to dictionaries.

    Parameters:
        model: The pydantic model or Python dictionary to be converted fully to a
            Python dictionary, through and through.
        **dict_kwargs (Dict[Any, Any]): Keyword arguments to be passed to `dict()`
            method calls for pydantic models.
            Note, this will be used for _all_ `dict()` method calls.

    Returns:
        A Python dictionary, where all nested values that were pydantic models are also
        converted to Python dictionaries.

    """

    def _internal(model_: "Any") -> "Any":
        """Internal function to be used recursively."""
        if isinstance(model_, dict):
            return {key: _internal(value) for key, value in model_.items()}
        if isinstance(model_, Iterable) and not isinstance(model_, (bytes, str)):
            return type(model_)(_internal(value) for value in model_)  # type: ignore[call-arg]  # pylint: disable=line-too-long
        if isinstance(model_, BaseModel):
            return _internal(model_.dict(**dict_kwargs))
        return model_

    if isinstance(model, BaseModel):
        res = model.dict(**dict_kwargs)
    elif isinstance(model, dict):
        res = deepcopy(model)
    else:
        raise TypeError("model must be either a pydantic model or a dict.")

    return _internal(res)



"""Custom "pydantic" types used in OTEAPI-OPTIMADE."""
import logging
import re
from typing import TYPE_CHECKING, cast, no_type_check
from urllib.parse import quote as urlquote
from urllib.parse import urlparse, urlunparse

from optimade.models import (
    EntryInfoResponse,
    EntryResponseMany,
    EntryResponseOne,
    InfoResponse,
    LinksResponse,
    ReferenceResponseMany,
    ReferenceResponseOne,
    StructureResponseMany,
    StructureResponseOne,
    Success,
)
from pydantic.networks import ascii_domain_regex, errors, int_domain_regex, url_regex
from pydantic.utils import update_not_none
from pydantic.validators import constr_length_validator, str_validator

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, Pattern, Tuple, TypedDict, Union

    from pydantic.config import BaseConfig
    from pydantic.fields import ModelField
    from pydantic.networks import CallableGenerator, Parts

    class OPTIMADEParts(TypedDict, total=False):
        """Similar to `pydantic.networks.Parts`."""

        base_url: str
        version: Optional[str]
        endpoint: Optional[str]
        query: Optional[str]


_OPTIMADE_BASE_URL_REGEX = None
_OPTIMADE_ENDPOINT_REGEX = None

LOGGER = logging.getLogger("oteapi_optimade.models")
LOGGER.setLevel(logging.DEBUG)


def optimade_base_url_regex() -> "Pattern[str]":
    """A regular expression for an OPTIMADE base URL."""
    global _OPTIMADE_BASE_URL_REGEX  # pylint: disable=global-statement
    if _OPTIMADE_BASE_URL_REGEX is None:
        _OPTIMADE_BASE_URL_REGEX = re.compile(
            r"^(?P<base_url>"
            # scheme https://tools.ietf.org/html/rfc3986#appendix-A
            r"(?:[a-z][a-z0-9+\-.]+://)?"
            r"(?:[^\s:/]*(?::[^\s/]*)?@)?"  # user info
            r"(?:"
            r"(?:\d{1,3}\.){3}\d{1,3}(?=$|[/:#?])|"  # ipv4
            r"\[[A-F0-9]*:[A-F0-9:]+\](?=$|[/:#?])|"  # ipv6
            r"[^\s/:?#]+"  # domain, validation occurs later
            r")?"
            r"(?::\d+)?"  # port
            r"(?P<path>/[^\s?#]*)?"  # path
            r")",
            re.IGNORECASE,
        )
    return _OPTIMADE_BASE_URL_REGEX


def optimade_endpoint_regex() -> "Pattern[str]":
    """A regular expression for an OPTIMADE base URL."""
    global _OPTIMADE_ENDPOINT_REGEX  # pylint: disable=global-statement
    if _OPTIMADE_ENDPOINT_REGEX is None:
        _OPTIMADE_ENDPOINT_REGEX = re.compile(
            # version
            r"(?:/(?P<version>v[0-9]+(?:\.[0-9+]){0,2})"
            r"(?=/info|/links|/version|/structures|/references|/calculations"
            r"|/extensions))?"
            # endpoint
            r"(?:/(?P<endpoint>(?:info|links|versions|structures|references"
            r"|calculations|extensions)(?:/[^\s?#]*)?))?$"
        )
    return _OPTIMADE_ENDPOINT_REGEX


class OPTIMADEUrl(str):
    """A deconstructed OPTIMADE URL.

    An OPTIMADE URL is made up in the following way:

        <BASE URL>/[<VERSION>/]<ENDPOINT>?<QUERY PARAMETERS>

    Where parts in square brackets (`[]`) are optional.
    """

    strip_whitespace = True
    min_length = 1
    # https://stackoverflow.com/questions/417142/what-is-the-maximum-length-of-a-url-in-different-browsers  # pylint: disable=line-too-long
    max_length = 2083
    allowed_schemes = {"http", "https"}
    tld_required = False
    user_required = False

    __slots__ = (
        "base_url",
        "version",
        "endpoint",
        "query",
        "scheme",
        "tld",
        "host_type",
    )

    @no_type_check
    def __new__(cls, url: "Optional[str]" = None, **kwargs) -> object:
        return str.__new__(
            cls,
            cls.build(**kwargs) if url is None else url,  # pylint: disable=missing-kwoa
        )

    def __init__(
        self,
        url: str,
        *,
        base_url: "Optional[str]" = None,
        version: "Optional[str]" = None,
        endpoint: "Optional[str]" = None,
        query: "Optional[str]" = None,
        scheme: "Optional[str]" = None,
        tld: "Optional[str]" = None,
        host_type: str = "domain",
    ) -> None:
        str.__init__(url)
        self.base_url = base_url
        self.version = version
        self.endpoint = endpoint
        self.query = query
        self.scheme = scheme
        self.tld = tld
        self.host_type = host_type

    @classmethod
    def build(
        cls,
        *,
        base_url: "str",
        version: "Optional[str]" = None,
        endpoint: "Optional[str]" = None,
        query: "Optional[str]" = None,
        **_kwargs: str,
    ) -> str:
        """Build complete URL from URL parts."""
        url = base_url.rstrip("/")
        if version:
            url += f"/{version}"
        if endpoint:
            url += f"/{endpoint}"
        if query:
            url += f"?{query}"
        return url

    @classmethod
    def __modify_schema__(cls, field_schema: "Dict[str, Any]") -> None:
        update_not_none(
            field_schema,
            minLength=cls.min_length,
            maxLength=cls.max_length,
            format="uri",
        )

    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        yield cls.validate

    @staticmethod
    def urlquote_qs(url: str) -> str:
        """Use `urllib.parse.quote` for query part of URL."""
        parsed_url = urlparse(url)
        quoted_query = urlquote(parsed_url.query, safe="=&,")
        parsed_url_list = list(parsed_url)
        parsed_url_list[-2] = quoted_query
        return urlunparse(parsed_url_list)

    @classmethod
    def validate(
        cls, value: "Any", field: "ModelField", config: "BaseConfig"
    ) -> "OPTIMADEUrl":
        """Pydantic validation of an OPTIMADE URL."""
        if value.__class__ == cls:
            return value

        value: str = str_validator(value)
        if cls.strip_whitespace:
            value = value.strip()
        url: str = cast(str, constr_length_validator(value, field, config))
        url = cls.urlquote_qs(url)

        url_match = url_regex().match(url)
        if url_match is None:
            raise ValueError(f"Cannot match URL ({url!r}) as a valid URL.")

        original_parts = cast("Parts", url_match.groupdict())
        parts = cls.apply_default_parts(original_parts)
        host, tld, host_type, rebuild = cls.validate_host(parts)
        optimade_parts = cls.build_optimade_parts(parts, host)
        optimade_parts = cls.validate_parts(parts, optimade_parts)

        if url_match.end() != len(url):
            raise errors.UrlExtraError(extra=url[url_match.end() :])

        return cls(
            None if rebuild else url,
            base_url=optimade_parts["base_url"],
            version=optimade_parts["version"],
            endpoint=optimade_parts["endpoint"],
            query=optimade_parts["query"],
            scheme=parts["scheme"],
            tld=tld,
            host_type=host_type,
        )

    @classmethod
    def validate_host(cls, parts: "Parts") -> "Tuple[str, Optional[str], str, bool]":
        """Validate host-part of the URL."""
        host: "Optional[str]" = None
        tld: "Optional[str]" = None
        rebuild: bool = False
        for host_type in ("domain", "ipv4", "ipv6"):
            host = parts[host_type]  # type: ignore[literal-required]
            if host:
                break
        else:
            raise errors.UrlHostError()

        if host_type == "domain":
            is_international = False
            domain = ascii_domain_regex().fullmatch(host)
            if domain is None:
                domain = int_domain_regex().fullmatch(host)
                if domain is None:
                    raise errors.UrlHostError()
                is_international = True

            tld = domain.group("tld")
            if tld is None and not is_international:
                domain = int_domain_regex().fullmatch(host)
                if domain is None:
                    raise ValueError("domain cannot be None")
                tld = domain.group("tld")
                is_international = True

            if tld is not None:
                tld = tld[1:]
            elif cls.tld_required:
                raise errors.UrlHostTldError()

            if is_international:
                host_type = "int_domain"
                rebuild = True
                host = host.encode("idna").decode("ascii")
                if tld is not None:
                    tld = tld.encode("idna").decode("ascii")

        return host, tld, host_type, rebuild

    @staticmethod
    def get_default_parts(parts: "Parts") -> "Parts":
        """Dictionary of default URL-part values."""
        return {"port": "80" if parts["scheme"] == "http" else "443"}

    @classmethod
    def apply_default_parts(cls, parts: "Parts") -> "Parts":
        """Apply default URL-part values if no value is given."""
        for key, value in cls.get_default_parts(parts).items():
            if not parts[key]:  # type: ignore[literal-required]
                parts[key] = value  # type: ignore[literal-required]
        return parts

    @classmethod
    def build_optimade_parts(cls, parts: "Parts", host: str) -> "OPTIMADEParts":
        """Convert URL parts to equivalent OPTIMADE URL parts."""
        base_url = f"{parts['scheme']}://"
        if parts["user"]:
            base_url += parts["user"]
        if parts["password"]:
            base_url += f":{parts['password']}"
        if parts["user"] or parts["password"]:
            base_url += "@"
        base_url += host
        # Hide port if it's a standard HTTP (80) or HTTPS (443) port.
        if parts["port"] and parts["port"] not in ("80", "443"):
            base_url += f":{parts['port']}"
        if parts["path"]:
            base_url += parts["path"]

        base_url_match = optimade_base_url_regex().fullmatch(base_url)
        LOGGER.debug(
            "OPTIMADE base URL regex match groups: %s",
            base_url_match.groupdict() if base_url_match else base_url_match,
        )
        if base_url_match is None:
            raise ValueError(
                "Could not match given string with OPTIMADE base URL regex."
            )

        endpoint_match = optimade_endpoint_regex().findall(
            base_url_match.group("path") if base_url_match.group("path") else ""
        )
        LOGGER.debug("OPTIMADE endpoint regex matches: %s", endpoint_match)
        for path_version, path_endpoint in endpoint_match:
            if path_endpoint:
                break
        else:
            LOGGER.debug("Could not match given string with OPTIMADE endpoint regex.")
            path_version, path_endpoint = "", ""

        base_url = base_url_match.group("base_url")
        if path_version:
            base_url = base_url[: -(len(path_version) + len(path_endpoint) + 2)]
        elif path_endpoint:
            base_url = base_url[: -(len(path_endpoint) + 1)]

        optimade_parts = {
            "base_url": base_url.rstrip("/"),
            "version": path_version or None,
            "endpoint": path_endpoint or None,
            "query": parts["query"],
        }
        return cast("OPTIMADEParts", optimade_parts)

    @classmethod
    def validate_parts(
        cls, parts: "Parts", optimade_parts: "OPTIMADEParts"
    ) -> "OPTIMADEParts":
        """
        A method used to validate parts of an URL.
        Could be overridden to set default values for parts if missing
        """
        scheme = parts["scheme"]
        if scheme is None:
            raise errors.UrlSchemeError()

        if cls.allowed_schemes and scheme.lower() not in cls.allowed_schemes:
            raise errors.UrlSchemePermittedError(set(cls.allowed_schemes))

        port = parts["port"]
        if port is not None and int(port) > 65_535:
            raise errors.UrlPortError()

        user = parts["user"]
        if cls.user_required and user is None:
            raise errors.UrlUserInfoError()

        base_url = optimade_parts["base_url"]
        if base_url is None:
            raise errors.UrlError()

        return optimade_parts

    def __repr__(self) -> str:
        extra = ", ".join(
            f"{n}={getattr(self, n)!r}"
            for n in self.__slots__
            if getattr(self, n) is not None
        )
        return f"{self.__class__.__name__}({super().__repr__()}, {extra})"

    def response_model(self) -> "Union[Tuple[Success, ...], Success, None]":
        """Return the endpoint's corresponding response model (from OPT)."""
        if not self.endpoint or self.endpoint == "versions":
            return None
        return {
            "info": (InfoResponse, EntryInfoResponse),
            "links": LinksResponse,
            "structures": (StructureResponseMany, StructureResponseOne),
            "references": (ReferenceResponseMany, ReferenceResponseOne),
            "calculations": (EntryResponseMany, EntryResponseOne),
        }.get(self.endpoint, Success)



#!/usr/bin/env python3
"""End-to-end testing in CI for OTEAPI OPTIMADE (using OTELib)."""
# pylint: disable=import-outside-toplevel
import importlib
import json
import os
import sys


def _check_import(package_: str) -> str:
    """Utility function to check a relevant Python package exists."""
    try:
        importlib.import_module(package_)
    except ImportError:
        return (
            f"{package_} not found on the system!\n"
            f"Please install {package_} by running:\n\n"
            f"    pip install {package_}\n\n"
        )
    else:
        return ""


def _check_service_availability(service_url: str) -> None:
    """Utility function to check availability of a web service.

    Note, this expects the service is created using FastAPI.
    Or that at least a reachable `/docs` endpoint exists.
    """
    import requests

    try:
        response = requests.get(f"{service_url}/docs", allow_redirects=True)
    except (requests.ConnectionError, requests.ConnectTimeout) as exc_:
        raise RuntimeError(f"Cannot connect to {service_url} !") from exc_
    else:
        if not response.ok:
            raise RuntimeError(f"Cannot connect to {service_url} !")


def main(oteapi_url: str) -> None:
    """The main test function.

    The test information is taken from here:
    https://github.com/Materials-Consortia/optimade-python-tools/blob/master/tests/server/query_params/test_filter.py#L52
    and here:
    https://github.com/Materials-Consortia/optimade-python-tools/blob/master/tests/server/query_params/test_filter.py#L391

    """
    from optimade.adapters import Structure
    from otelib import OTEClient
    from pydantic import ValidationError

    from oteapi_optimade.models import OPTIMADEResourceSession

    client = OTEClient(oteapi_url)

    config = {
        "query_parameters": {
            "filter": 'elements HAS "Ac"',
            "page_limit": 2,
        }
    }

    source = client.create_dataresource(
        accessService="OPTIMADE",
        accessUrl=f"http://localhost:{os.getenv('OPTIMADE_PORT', '5000')}/",
        configuration=config,
    )

    session = source.get()

    try:
        session = OPTIMADEResourceSession(**json.loads(session))
    except ValidationError as exc_:
        raise RuntimeError(
            "Could not parse returned session as an OPTIMADEResourceStrategy."
        ) from exc_

    assert session.optimade_resource_model == f"{Structure.__module__}:Structure"
    assert len(session.optimade_resources) == 2

    for resource in tuple(session.optimade_resources):
        parsed_resource = Structure(resource)
        assert parsed_resource.id in ["mpf_1", "mpf_110"]

    # Use a filter strategy
    query = client.create_filter(
        filterType="OPTIMADE",
        query='NOT( elements HAS "Ag" AND nelements>1 )',
        limit=4,
        configuration=config,
    )

    pipeline = query >> source
    session = pipeline.get()

    try:
        # Should be an OPTIMADEResourceSession because `source` is last in the pipeline
        session = OPTIMADEResourceSession(**json.loads(session))
    except ValidationError as exc_:
        raise RuntimeError(
            "Could not parse returned session as an OPTIMADEResourceStrategy."
        ) from exc_

    assert session.optimade_resource_model == f"{Structure.__module__}:Structure"
    assert len(session.optimade_resources) == 4

    for resource in tuple(session.optimade_resources):
        parsed_resource = Structure(resource)
        assert parsed_resource.id in [
            "mpf_1",
            "mpf_23",
            "mpf_30",
            "mpf_110",
            "mpf_3803",
            "mpf_3819",
            "mpf_200",
        ]


if __name__ == "__main__":
    # Check Python packages
    NECESSARY_PACKAGES = "otelib", "oteapi_optimade", "requests", "optimade"
    MESSAGES = ""
    for package in NECESSARY_PACKAGES:
        MESSAGES += _check_import(package)
    if MESSAGES:
        raise ImportError(MESSAGES)

    # Configuration
    PORT = os.getenv("OTEAPI_PORT", "8080")
    OTEAPI_SERVICE_URL = f"http://localhost:{PORT}"
    OTEAPI_PREFIX = os.getenv("OTEAPI_prefix", "/api/v1")
    if "OTEAPI_prefix" not in os.environ:
        # Set environment variables
        os.environ["OTEAPI_prefix"] = OTEAPI_PREFIX

    try:
        _check_service_availability(service_url=OTEAPI_SERVICE_URL)
    except RuntimeError as exc:
        sys.exit(exc)

    try:
        main(oteapi_url=OTEAPI_SERVICE_URL)
    except Exception as exc:  # pylint: disable=broad-except
        sys.exit(exc)



"""`oteapi_optimade.models` module - pydantic data models."""
from .filter import OPTIMADEFilterConfig, OPTIMADEFilterSession
from .parse import OPTIMADEParseConfig, OPTIMADEParseSession
from .resource import OPTIMADEResourceConfig, OPTIMADEResourceSession

__all__ = (
    "OPTIMADEFilterConfig",
    "OPTIMADEFilterSession",
    "OPTIMADEParseConfig",
    "OPTIMADEParseSession",
    "OPTIMADEResourceConfig",
    "OPTIMADEResourceSession",
)
